#!/usr/bin/env python2.7

from argparse import ArgumentParser
from numpy import arange, array, concatenate, diff, nonzero, setdiff1d, sqrt, uint16, zeros
from numpy.random import randint, randn
from scipy.special import erfinv
from struct import pack, unpack
import sys

# PLSR file is written as:
#   <fileheader><pkt><pkt>...<pkt>
# <fileheader> = <num_chan><num_samples><num_bits><num_packets>
# where
#   num_chan = number of channels (uint32_t)
#   num_samples = number of samples in all channels per VDIF packet (uint32_t)
#   num_bits = number of bits per sample (uint32_t)
#   num_packets = number of packets in file (uint32_t) !!!including invalids, in the original VDIF!!!
# <pkt> = <pkt_hdr><pkt_dat>
# <pkt_hdr> = <sec><cnt>
# where
#   sec = seconds from reference epoch as in original VDIF (uint32_t)
#   cnt = data frame number within second as in original VDIF (uint32_t)
# <pkt_dat> = <ch0s00><ch0s01><ch0s10><ch1s00>...<chMs10>
# where
#   ch0s00 = count channel 0 samples in packet equal to 0b00 (uint16_t)
#   ch0s01 = count channel 0 samples in packet equal to 0b01 (uint16_t)
#   ch0s10 = count channel 0 samples in packet equal to 0b10 (uint16_t)
#   ch1s00 = count channel 1 samples in packet equal to 0b00 (uint16_t)
#   ...
#   chMs10 = count channel M samples in packet equal to 0b10 (uint16_t)

def read_plsr_file_header(fh):
	fh.seek(0,0)
	return unpack("<4I",fh.read(16))

def vdif_frame_step(sec0,df0,sec,df,fps=125000):
	t0 = sec0*fps + df0
	t = sec*fps + df
	return t - t0

def fill(fh,n,nchannel,nbit,outfile):
	num_states = 2**nbit-1
	count_per_packet = num_states*nchannel
	#             header + sizeof(uint16_t)*num_counts
	bytes_per_packet = 8 + 2*count_per_packet
	print "{0} b/pkt x {1} pkt".format(bytes_per_packet,n)
	# keep track of timestamp in case packets missing
	sec_prev = None
	df_prev = None
	# create initial filler buffer, assume threshold at 1-sigma
	n_samples_per_packet = 1000
	n_buf = 100
	state_thresholds = [-1,0,1]
	r = randn(n_samples_per_packet,nchannel,n_buf)
	buf = zeros((n_buf,nchannel,num_states),dtype=uint16)
	for ii in range(n_buf):
		for cc in range(nchannel):
			for ss in range(num_states):
				buf[ii,cc,ss] = nonzero(r[:,cc,ii]<state_thresholds[ss])[0].size - buf[ii,cc,:ss].sum()
	# open output file
	with open(outfile,"w") as fh_out:
		# copy header to output file
		fh_out.write(pack("<4I",nchannel,nsample,nbit,n))
		# reduce each packet
		n_update = n/20
		for ii in range(n):
			if (ii % n_update) + 1 == n_update:
				print "{0:.2f}%".format(ii*100.0/n)
			b = fh.read(bytes_per_packet)
			if len(b) < bytes_per_packet:
				print "WARNING: short read or eof after {0} packets (expected {1}), breaking...".format(ii,n)
				break
			sec,df = unpack("<2I",b[:8])
			if sec_prev is not None and df_prev is not None:
				step = vdif_frame_step(sec_prev,df_prev,sec,df)
			else:
				print "starting at {0}.{1:06d}".format(sec,df)
				step = 0
			if step > 1:
				print "WARNING: step to {0}.{1:06d} is {2}".format(sec,df,step)
				# generate filler data
				filler_data = zeros((step-1,nchannel,num_states))
				for cc in range(nchannel):
					for ss in range(num_states):
						filler_data[:,cc,ss] = randint(buf[:,cc,ss].min(),buf[:,cc,ss].max()+1,size=step-1)
				fake_df = df_prev
				fake_sec = sec_prev
				for jj in range(step-1):
					fake_counts = filler_data[jj,:,:].flatten()
					fake_df = fake_df + 1
					if fake_df >= 125000:
						fake_sec = fake_sec + 1
						fake_df = 0
					# write to file
					fh_out.write(pack("<II{0}H".format(nchannel*num_states),fake_sec,fake_df,*fake_counts))
				print "filled with random data in range ( {0}.{1:06d} , {2}.{3:06d} ]".format(sec_prev,df_prev,fake_sec,fake_df)
			# after fill, continue here
			counts = array(unpack("<{0}H".format(count_per_packet),b[8:]),dtype=uint16).reshape((-1,3))
			# update filler buffer, only with actual data
			buf[ii % n_buf,:,:] = counts
			# continue writing to file
			fh_out.write(b)
			# update timestamp irrespective of filler or actual data
			sec_prev = sec
			df_prev = df

if __name__ == "__main__":
	ap = ArgumentParser(description="Copy PLSR file and fill missing data")
	ap.add_argument("-n","--num-pkt",metavar="NPKT",type=int,default=-1,
	  help="read NPKT number of packets only, or all if <0 (default is -1)")
	ap.add_argument("-o","--output",metavar="OUTFILE",type=str,default="tmp.rdx1",
	  help="write output to OUTFILE (default is 'tmp.rdx1')")
	ap.add_argument("infile",metavar="INPUT",type=str,
	  help="read data from INPUT")
	args = ap.parse_args()
	
	with open(args.infile,"r") as fh:
		nchannel,nsample,nbit,npkt = read_plsr_file_header(fh)
		
		if args.num_pkt >= 0:
			N = args.num_pkt
		else:
			N = npkt
		
		print "Converting {0} to {1}".format(args.infile,args.output)
		c = fill(fh,N,nchannel,nbit,args.output)

