#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "presto2.h"
#include "vectors2.h"
#include "main_cmd.h"
#include "psrfits.h"
#include "vdifio.h"

void rewind(FILE *f);
int check_vals(int bytes_per_rec, double dt, FILE * infile);
void coord_double2str(double x, char *outstr, int is_ra);

int verbose = 1;

int main(int argc, char *argv[])
{
  FILE *infile;
  unsigned char *rawdata, *tmprec, *btmp;
  char basename[80], source_name[80], str_ra[20], str_dec[20];
  double dchan, dt, offs;
  float fch1;
  float *freqs, *ftmp1;
  int nchan, bytes_per_rec, bytes_per_block, nsblk, vals_read;
  int bytes_per_subint, spec_per_row;
  int jd, sec, usec;
  long imjd, smjd;
  double mjd_f, smjd_f;
  int ii, jj, status;
  int rownum, breakout;
  struct psrfits pf;
  Cmdline *cmd;

  /* Call usage() if we have no command line arguments */
  if (argc == 1) {
    Program = argv[0];
    usage();
    exit(0);
  }

  /* Parse the command line using the excellent program Clig */
  cmd = parseCmdline(argc, argv);

  strcpy(basename, cmd->outfile);
  nchan = cmd->nchan;
  dchan = cmd->dchan;
  dt    = cmd->dt / 1000.0 / 1000.0;

  /* Open file and check first two records */
  infile = chkfopen(cmd->argv[0], "rb");

  int framepersec = 1.0/dt; // VDIF frames are 8 microseconds
  fprintf(stderr,"Frames per second = %d\n",framepersec);
  fprintf(stderr,"Channels = %d\n",nchan);

  //~ State counts file no longer starts with VDIF header, the second in
  //~ epoch, and data frame in second are now written as uint32_t in 
  //~ the first packet header.
  //~ //Read the VDIF Header
  //~ struct vdif_header VDIF_HEADER;
  //~ fread(&VDIF_HEADER, sizeof(VDIF_HEADER), 1, infile);  
  //~ double VDIF_mjd = getVDIFDMJD(&VDIF_HEADER, framepersec); 
  //~ .....calculate VDIF_mjd here instead:
  /* file format is now:
# The data file is written as:
#   <fileheader><pkt><pkt>...<pkt>
# <fileheader> = <num_chan><num_samples><num_bits><num_packets>
# where
#   num_chan = number of channels (uint32_t)
#   num_samples = number of samples in all channels per VDIF packet (uint32_t)
#   num_bits = number of bits per sample (uint32_t)
#   num_packets = number of packets in file (uint32_t) !!!including invalids, in the original VDIF!!!
# <pkt> = <pkt_hdr><pkt_dat>
# <pkt_hdr> = <sec><cnt>
# where
#   sec = seconds from reference epoch as in original VDIF (uint32_t)
#   cnt = data frame number within second as in original VDIF (uint32_t)
    */
  uint32_t file_header[4];
  uint32_t vdif_seconds;
  uint32_t vdif_dataframe;
  fread((void *)file_header, sizeof(uint32_t)*4, 1, infile);
  fread((void *)&vdif_seconds, sizeof(uint32_t), 1, infile);
  fread((void *)&vdif_dataframe, sizeof(uint32_t), 1, infile);
  int epoch = 34;
  double VDIF_mjd = (double)ymd2mjd(2000 + epoch/2, (epoch%2)*6+1, 1) +
    ((double)vdif_seconds)/86400.0 + ((double)vdif_dataframe)/86400.0/125000.;
  fprintf(stderr,"VDIF_mjd = %f\n",VDIF_mjd);
  
  /* Set center frequencies of channels */
  freqs   = gen_fvect(nchan);

  for(ii=0; ii<nchan; ii++){
    freqs[nchan-ii-1] = 86268 - (0.5 + ii) * dchan;
  }
  

  /* Set up STT PSRFITS Header Values */
  mjd_f = VDIF_mjd; //jd - 2400000.5;
  imjd = (long) mjd_f;
  smjd_f = (mjd_f - imjd) * 24 * 3600;
  smjd = (long) smjd_f;
  offs = usec * 1E-6 + (smjd_f - smjd);

  coord_double2str(cmd->raj, str_ra, 1);   // is_ra = 1
  coord_double2str(cmd->dej, str_dec, 0);  // is_ra = 0
  strcpy(source_name, cmd->src_name);

  printf("First record values: \n");
  printf("  STT_IMJD = %ld\n", imjd);
  printf("  STT_SMJD = %ld\n", smjd);
  printf("  STT_OFFS = %.6f\n", offs);

  printf("%d freq channels from %.1f to %.1f in steps of %.1f:  BW = %.1f\n",
	 nchan, freqs[0], freqs[nchan-1],
	 dchan, freqs[nchan-1] - freqs[0] + dchan);

  printf("RA  = %s \n", str_ra);
  printf("DEC = %s \n", str_dec);
  printf("Source = %s \n", source_name);

  fill_psrfits_struct(32, &pf, dt, source_name, nchan, dchan,
                      imjd, smjd, offs, str_ra, str_dec, freqs[0], basename);
  

  /* format per record <pkt> is:
# <pkt> = <pkt_hdr><pkt_dat>
# <pkt_hdr> = <sec><cnt>
# where
#   sec = seconds from reference epoch as in original VDIF (uint32_t)
#   cnt = data frame number within second as in original VDIF (uint32_t)
# <pkt_dat> = <ch0s00><ch0s01><ch0s10><ch1s00>...<chMs10>
# where
#   ch0s00 = count channel 0 samples in packet equal to 0b00 (uint16_t)
#   ch0s01 = count channel 0 samples in packet equal to 0b01 (uint16_t)
#   ch0s10 = count channel 0 samples in packet equal to 0b10 (uint16_t)
#   ch1s00 = count channel 1 samples in packet equal to 0b00 (uint16_t)
#   ...
#   chMs10 = count channel M samples in packet equal to 0b10 (uint16_t)
   *   so the bytes-per-record is 2*sizeof(uint32_t)+nchan*3*sizeof(uint16_t)
   */
  bytes_per_rec = (int)(2*sizeof(uint32_t) + (nchan) * 3 * sizeof(uint16_t));
  
  spec_per_row = pf.hdr.nsblk; // this is the number of samples per second
  bytes_per_subint = pf.sub.bytes_per_subint;
  pf.multifile = 1;
  pf.rows_per_file = (int) ((cmd->outlenMB * 1048576.0) / 
			    (double) pf.sub.bytes_per_subint);
  
  status = psrfits_create(&pf);
  if (status) {
    printf("\nError (%d) creating PSRFITS...\n\n", status);
    status = 0;
  }

  /* Allocate and initialize a raw data and temp array */
  bytes_per_block = bytes_per_rec * spec_per_row;
  
  rawdata = gen_bvect(bytes_per_block);
  for(ii=0; ii < bytes_per_block; ii++){
    rawdata[ii] = (unsigned char) 0;
  }

  btmp = gen_bvect(bytes_per_subint);
  for(ii=0; ii < bytes_per_subint; ii++){
    btmp[ii] = (unsigned char) 0;
  }

  /* Allocate and initialize two temporary float arrays */
  //uint16_t uitmp1_array[bytes_per_block / 2];
  //uint16_t *uitmp1 = uitmp1_array;
  uint16_t *uitmp1 = (uint16_t *)malloc(bytes_per_block);
  for(ii=0; ii < bytes_per_block / 2; ii++){
    uitmp1[ii] = 0.0;
  }

  /* Allocate and initialize two temporary float arrays */
  ftmp1 = gen_fvect(bytes_per_block / 4);
  for(ii=0; ii < bytes_per_block / 4; ii++){
    ftmp1[ii] = 0.0;
  }

  /* Initialize subint data  */
  for(ii=0; ii < bytes_per_subint; ii++){
    pf.sub.data[ii] = (unsigned char) 0;
  }

  /* Set up the frequency psrfits data */
  for(ii=0; ii < nchan; ii++){
    pf.sub.dat_freqs[ii] = freqs[ii];
  }

  rownum = 0;
  breakout = 0;
  
  fprintf(stderr,"bytes_per_subint: %d\n",bytes_per_subint);
  fprintf(stderr,"pf.sub.bytes_per_subint: %d\n",pf.sub.bytes_per_subint);
  fprintf(stderr,"bytes_per_block: %d\n",bytes_per_block);
  fflush(stderr);
  
  //Constants for total power estimation from state counts
  double x0 = 1.0;
  double G0 = 0.241971;
  double p0 = 0.682689;

  /* Read in data for one scan by row */
  while(1){
    rownum++;
    //printf("\rWorking on row %d\r", rownum);
    //fflush(stdout);
    
    /* Read in one row worth of data into rawdata buffer */
    vals_read = fread(rawdata, sizeof(unsigned char), bytes_per_block, infile);

    //if (verbose) fprintf(stderr,"vals_read: %d\nState Counts: %d\n",vals_read,state_counts);

    if(vals_read < bytes_per_block){
      printf("\n\nRead %d vals (expected %d)\n", vals_read, bytes_per_block);
      breakout = 1;
    }

    if(breakout == 1){
      break;
    }

    /* Cast rawdata to temporary unsigned int array */
    uitmp1 = (uint16_t *) rawdata;
    //fprintf(stderr,"%d\n",uitmp1[0]);

    /* Compute the power (and reverse things) */
    for(ii=0; ii < spec_per_row; ii++){
		
      for(jj=0; jj < nchan; jj++){
		  /* offset into uitmp1 is:
		   *     ii*bytes_per_record*sizeof(char) = ii*bytes_per_record/2*sizeof(uint16_t)
		   *   + 2*sizeof(uint32_t) = 4*sizeof(uint16_t) <-- per-packet timestamp
		   *   + jj*3*sizeof(uint16_t) <-- there are 3 state counts per channel
		   *   + 1*sizeof(uint16_t) <-- this is count for state0
		   */
		uint16_t s12 = uitmp1[ii*bytes_per_rec/2 + jj*3 + 1]; // + 4 left out, because we already read the first one to calculate MJD
		// s12 is now count(state1), add count(state2)
		s12 += uitmp1[ii*bytes_per_rec/2 + jj*3 + 2]; // + 4 left out, because we already read the first one to calculate MJD
        double p = (double)s12/1000.0; // <-- there are state counts per 1000 samples in each packet 
	    ftmp1[ii * nchan + jj] = 1.0 + (p0 - p)/(x0*G0);
      }
    }

    /* Cast ftmp1 into temporary binary array */
    btmp = (unsigned char *) ftmp1;
    
    /* Fill subint data buffer */
    for(ii=0; ii<bytes_per_subint; ii++){
      pf.sub.data[ii] = btmp[ii];
    }

    /* Now write the subint row */
    pf.sub.offs = (pf.tot_rows + 0.5) * pf.sub.tsubint;
    //fprintf(stderr,"pf.tot_rows: %d\n",pf.tot_rows);
    //fprintf(stderr,"pf.rownum: %d\n",pf.rownum);
    //fprintf(stderr,"pf.rows_per_file: %d\n",pf.rows_per_file);

    status = psrfits_write_subint(&pf);
    if (status) {
      printf("\nError (%d) writing PSRFITS...\n\n", status);
      break;
    }
  }
  /* Free up everything */
  free(rawdata);
  free(freqs);

  free(pf.sub.data);
  //free(ftmp1);
  //free(ftmp2);
  //free(btmp);

  return 0;

}


/* Check file consistency (Now depreciated) */
int check_vals(int bytes_per_rec, double dt, FILE * infile){
  int retval, vals_read, j1, j2, s1, s2, u1, u2;
  float fch1, fch2, dt2;
  unsigned char *tmprec;

  retval = 1;
  tmprec  = gen_bvect(2 * bytes_per_rec);

  /* Make sure we're at the beginning */
  rewind(infile);

  /* Read in two records worth of data */
  vals_read = fread(tmprec, 1, 2 * bytes_per_rec, infile);
  if(vals_read != 2 * bytes_per_rec){
    printf("Read %d vals from first record (expected %d)\n", vals_read, bytes_per_rec);
    return 0;
  }

  /* Set Values */
  memcpy(&j1, &tmprec[0], sizeof(int));
  memcpy(&s1, &tmprec[4], sizeof(int));
  memcpy(&u1, &tmprec[8], sizeof(int));
  memcpy(&fch1, &tmprec[12], sizeof(float));

  memcpy(&j2, &tmprec[0 + bytes_per_rec], sizeof(int));
  memcpy(&s2, &tmprec[4 + bytes_per_rec], sizeof(int));
  memcpy(&u2, &tmprec[8 + bytes_per_rec], sizeof(int));
  memcpy(&fch2, &tmprec[12 + bytes_per_rec], sizeof(float));

  /* Compare Values */
  dt2 = ((j2-j1) * 24 * 3600 + (s2-s1)) * 1E6 + (u2-u1);
  if((dt2 - dt * 1E6) > 1){
    printf("dt2 = %.1f us, dt = %.1f us\n", dt2, dt * 1E6);
    retval *= 0;
  }

  if(fch1 != fch2){
    printf("fch2 = %.2f, fch1 = %.2f\n", fch2, fch1);
    retval *= 0;
  }

  /* Make sure we're at the beginning */
  rewind(infile);

  free(tmprec);

  return retval;
}

/* Convert raj/dej double to string */
void coord_double2str(double x, char *outstr, int is_ra)
{
  int hh, mm;
  float ss;

  hh = (int) (x / 10000.0);
  mm = (int) (fabs(x - hh * 10000) / 100.0);
  ss = fabs(x - hh * 10000) - mm * 100;
  if(is_ra) // Print RA string if is_ra
    sprintf(outstr, "%.2i:%.2i:%07.4f", hh, mm, ss);
  else     // Else pring DEC string
    sprintf(outstr, "%+.2i:%.2i:%07.4f", hh, mm, ss);
}

