#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "vdif_files.h"
#include "vdif_frames.h"

#define PACKET_BATCH_SIZE 2000

uint32_t *create_states_table_32ch_2b() {
	uint32_t *table;
	
	table = (uint32_t *)malloc(32*4*sizeof(uint32_t));
	return table;
}

void destroy_states_table_32ch_2b(uint32_t *table) {
	free(table);
}

void reset_states_table_32ch_2b(uint32_t *table) {
	int ch;
	int st;
	
	for (ch=0; ch<32; ch++) {
		for (st=0; st<4; st++) {
			table[ch*4 + st] = 0;
		}
	}
}

void update_states_table_32ch_2b(uint32_t *table, int nsamp, uint32_t *updata) {
	int ii;
	int ch = 0;
	
	for (ii=0; ii<nsamp; ii++) {
		table[ch*4 + updata[ii]]++;
		ch = (ch + 1) % 32;
	}
}

void print_states_table_32ch_2b(uint32_t *table) {
	int ch;
	int st;
	
	fprintf(stdout,"# Ch            00b          01b          02b          03b\n");
	for (ch=0; ch<32; ch++) {
		fprintf(stdout,"%02d    ",ch);
		for (st=0; st<4; st++) {
			fprintf(stdout,"   %10d",table[ch*4+st]);
		}
		fprintf(stdout,"\n");
	}
}

void write_states_table_32ch_2b(int fd, uint32_t *table) {
	int ii = 0;
	int ch;
	int rv;
	int st;
	uint16_t out[96];
	
	for (ch=0; ch<32; ch++) {
		// only write counts for 3 states, 4th is redundant
		for (st=0; st<3; st++) {
			out[ii++] = table[ch*4 + st];
		}
	}
	rv = write(fd, out, 96*sizeof(uint16_t));
	if (rv < 0) {
		perror("write_states_table_32ch_2b");
	}
}

void write_outfile_header(int fd, int nchannel, int nsample, int nbit, int npacket) {
	int rv;
	uint32_t hdr[4];
	
	hdr[0] = nchannel;
	hdr[1] = nsample;
	hdr[2] = nbit;
	hdr[3] = npacket;
	rv = write(fd, hdr, 4*sizeof(uint32_t));
	if (rv < 0) {
		perror("write_outfile_header");
	}
}

void write_outpacket_header(int fd, vdif_header_t *pkt) {
	int rv;
	uint32_t hdr[2];
	
	hdr[0] = pkt->secs_since_epoch;
	hdr[1] = pkt->data_frame;
	rv = write(fd, hdr, 2*sizeof(uint32_t));
	if (rv < 0) {
		perror("write_outpacket_header");
	}
}

int main(int argc, const char **argv) {
	// generic counters
	int ii,jj,kk;
	// packets and properties
	FFile_t file;
	void *buf;
	int num_packets;
	int packets_read;
	int pkt_size_8byte;
	vdif_header_t *pkt;
	// sample data and formats
	uint32_t *data;
	int bps;
	int nch;
	int cmp;
	int num_samples;
	// time-keeping
	int ts_curr, tp_curr, ts_prev, tp_prev;
	// output
	int fd_out;
	int out_file_header;
	uint32_t *table;
	
	// process command line arguments
	if (argc < 3) {
		fprintf(stdout,"Usage: %s NUMPACKETS FLATFILE OUTFILE\n",&argv[0][2]);
		fprintf(stdout,"where NUMPACKETS is the number of VDIF frames to read, use -1 for all in file\n.");
		return 1;
	}
	num_packets = atoi(argv[1]);
	
	// initialize
	table = create_states_table_32ch_2b();
	if (open_file_f(argv[2],&file) == -1) {
		return 2;
	}
	fd_out = open(argv[3],O_CREAT|O_WRONLY|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
	if (fd_out == -1) {
		return 3;
	}
	out_file_header = 0;
	
	packets_read = 0;
	while (num_packets == -1 || packets_read<num_packets) {
		if (num_packets != -1 && num_packets-packets_read<PACKET_BATCH_SIZE) {
			ii = read_packets_from_file_f(&file,num_packets-packets_read,&buf);
			packets_read += ii;
		} else {
			ii = read_packets_from_file_f(&file,PACKET_BATCH_SIZE,&buf);
			packets_read += ii;
		}
		if (ii==0) {
			// this means end-of-file reached
			break;
		}
		pkt_size_8byte = ((vdif_header_t *)buf)->frame_length;
		for (jj=0; jj<ii; jj++) {
			pkt = (vdif_header_t *)(buf + jj*pkt_size_8byte*8);
			num_samples = get_samples(pkt,&data,&nch,&bps,&cmp);
			if (!out_file_header) {
				write_outfile_header(fd_out,nch,num_samples,bps,0);
				out_file_header = 1;
			}
			write_outpacket_header(fd_out,pkt);
			reset_states_table_32ch_2b(table);
			update_states_table_32ch_2b(table,num_samples,data);
			write_states_table_32ch_2b(fd_out,table);
			free(data);
		}
		free(buf);
	}
	// finish writing output file header
	lseek(fd_out, 0, SEEK_SET);
	write_outfile_header(fd_out,nch,num_samples,bps,packets_read);
	
	// deinitialize
	destroy_states_table_32ch_2b(table);
	close_file_f(&file);
	close(fd_out);
	
	return 0;
}
